package Timtosh::Complete;

use strict;
use vars qw(@ISA @EXPORT @EXPORT_OK);

require Exporter;
use Timtosh::Parser;
use Timtosh::Utils qw(escape);

@ISA = qw(Exporter);
@EXPORT = qw(complete);
@EXPORT_OK = ();

my %command_specific;

sub complete {
  my ($string, $loc) = @_;

  #print "\n";

  my ($index, @tokens) = tokenize($string, $loc);
  if (substr($string, $loc, 1) =~ /\s/ &&
      substr($string, $loc - 1, 1) =~ /[^\s]/) {
    $index--;
  }

  #print "active token: ", join(", ", @{$tokens[$index]}), "\n";

  my ($pre, $post);
  for (my $i = 0; $i < $index; $i++) {
    $pre .= $tokens[$i]->[0];
  }
  for (my $i = $index + 1; defined($tokens[$i]); $i++) {
    $post .= $tokens[$i]->[0];
  }

  my $replacement = do_completion($index, @tokens);
  #print "replacement: $replacement\n";

  if (defined($replacement)) {
    $string = $pre. $replacement;
    $loc = length($string);
    $string .= $post;
  }

  return ($string, $loc);
}

sub do_completion {
  my $index = shift;

  return undef unless $_[$index]->[1] eq 'WORD';

  my @tokens = @_;
  my $token = $tokens[$index];

  if ($index == 0) {
    return complete_command($index, $token->[2]);

  } else {
    my @files = Timtosh::Parser::expand_glob($token->[2]);
    if ($files[0] ne $token->[2]) {
      return join(" ", escape(@files));
    }

    my @list;
    my $new_index = $index;
    foreach (@tokens) {
      my @expansion = expand($_);
      if ($index-- >= 0) {
	$new_index += $#expansion;
      }
      foreach (@expansion) {
	push @list, $_->[1];
      }
    }
    $index = $new_index;

    my $cmd = $tokens[0]->[2];
    my $function;
    if (defined($command_specific{$cmd})) {
      $function = $command_specific{$cmd};
    } else {
      $function = \&complete_file;
    }

    return join(" ", &$function($index, @list));
  }
}

sub complete_file {
  my $index = shift;
  my @list = @_;

  my $word = $list[$index];

  my @candidates = Timtosh::Parser::expand_glob("$word*");

  if ($#candidates == 0 && $candidates[0] ne "$word*") {
    my $result;
    if (-d $candidates[0]) {
      $result = (escape("$candidates[0]/"))[0];
    } else {
      $result = (escape("$candidates[0]"))[0]. " ";
    }
    return $result;
  } elsif ($#candidates > 0) {
    return (escape(complete_from_list($word, @candidates)))[0];
  }

  return $word;
}

sub complete_command {
  my $index = shift;
  my @list = @_;
  my $word = $list[$index];
  my @candidates;

  if ($word =~ m:^\.*/: || $ENV{'PATH'} =~ /(^|:).(:|$)/) {
    # filter out non-executable stuff
    @candidates = grep { -x $_ } Timtosh::Parser::expand_glob("$word*");
  }
  push @candidates, grep(/^$word/, keys %::command);

  if ($#candidates == 0) {
    my $result;
    if (-d $candidates[0]) {
      $result = (escape("$candidates[0]/"))[0];
    } else {
      $result = (escape("$candidates[0]"))[0]. " ";
    }
    return $result;
  }

  return (escape(complete_from_list($word, @candidates)))[0];
}

sub complete_from_list {
  my $word = shift;
  my @list = @_;

  my @candidates = grep(/^$word/, @list);

  # lots of possible candidates
  my $matching = length($candidates[0]);
  for (my $i = 1; $i <= $#candidates; $i++) {
    my $j;
    for ($j = 0; $j < $matching; $j++) {
      last if(substr($candidates[$i - 1], $j, 1) ne
	  substr($candidates[$i], $j, 1));
    }
    $matching = $j;
  }

  return substr($candidates[0], 0, $matching);
}

1;
